
#include "ntddk.h"
#include "stdarg.h"
#include "stdio.h"
#include "ffsd.h"


BOOLEAN			bpxEnabled = FALSE;	// Enable/Disable IRP_MJ_READ bpx	
PDRIVER_OBJECT  FFsdDriver;			// FFSD driver object
PDEVICE_OBJECT  DriveHookDevice;	// Device object of the hooked drive


//*****************************************************************************
// 
// FFsdHooked (we don't use a completion routine)
//
//*****************************************************************************
NTSTATUS 
FFsdHooked( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp,
    IN PVOID Context 
    )
{
    
	IO_STATUS_BLOCK		 IoStatusBlock;

    if(Irp->PendingReturned) {

        IoMarkIrpPending(Irp);
    }
    return Irp->IoStatus.Status;
}



//****************************************************************************
//
// FFsdQueryFileComplete - Completion routine for our IRP_MJ_QUERY_INFORMATION
//
//****************************************************************************
NTSTATUS 
FFsdQueryFileComplete(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp,
    PVOID Context
    )
{
    *Irp->UserIosb = Irp->IoStatus;
    KeSetEvent(Irp->UserEvent,0,FALSE);		// Wakeup FFsdQueryFile routine
											// so we can look at the IRP results
    IoFreeIrp(Irp);
    return STATUS_MORE_PROCESSING_REQUIRED;
}



//****************************************************************************
//
// FFsdQueryFile - Build our IRP_MJ_QUERY_INFORMATION ,send down and wait the
//				   the completion routine to wakeup us again :)
//
//****************************************************************************
BOOLEAN 
FFsdQueryFile( 
    PDEVICE_OBJECT DeviceObject, 
    PFILE_OBJECT FileObject,
    FILE_INFORMATION_CLASS FileInformationClass,
    PVOID FileQueryBuffer,
    ULONG FileQueryBufferLength,
	PVOID CompletionRoutine,
	ULONG IRP_TYPE
    )
{
    PIRP irp;
    KEVENT event;
    IO_STATUS_BLOCK IoStatusBlock;
    PIO_STACK_LOCATION ioStackLocation;

    KeInitializeEvent(&event, SynchronizationEvent, FALSE);

    irp = IoAllocateIrp(DeviceObject->StackSize, FALSE);
    if(!irp) {

       return FALSE;
    }
  
    irp->AssociatedIrp.SystemBuffer = FileQueryBuffer;
    irp->UserEvent = &event;
    irp->UserIosb = &IoStatusBlock;
    irp->Tail.Overlay.Thread = PsGetCurrentThread();
    irp->Tail.Overlay.OriginalFileObject = FileObject;
    irp->RequestorMode = KernelMode;
    irp->Flags = 0;

    ioStackLocation = IoGetNextIrpStackLocation(irp);
    ioStackLocation->MajorFunction = IRP_MJ_QUERY_INFORMATION;
    ioStackLocation->DeviceObject = DeviceObject;
    ioStackLocation->FileObject = FileObject;
    ioStackLocation->Parameters.QueryFile.Length = FileQueryBufferLength;
    ioStackLocation->Parameters.QueryFile.FileInformationClass = FileInformationClass;

    IoSetCompletionRoutine(irp, CompletionRoutine, 0, TRUE, TRUE, TRUE);

    (void) IoCallDriver(DeviceObject, irp);		// send our IRP to the next (low) fsd driver

    KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0);

    return NT_SUCCESS(IoStatusBlock.Status);
}




//***************************************************************************
//
// FFsdGetPathName - We ask the filename for this IRP_MJ_READ hooked and then
//					 give a look at the parameters (bpx enabled)
//
//***************************************************************************
BOOLEAN 
FFsdGetPathName( 
    BOOLEAN IsFastIO, 
    PFILE_OBJECT fileObject, 
    PHOOK_EXTENSION hookExt, 
    PIO_STACK_LOCATION curParameters,
	PIRP  irp,
	ULONG ReadOffsetFastIO,
	ULONG Length,
	ULONG Buffer
    )
{
    ANSI_STRING         fileName;
    PFILE_NAME_INFORMATION fileNameInfo;
    FILE_INTERNAL_INFORMATION fileInternalInfo;
    UNICODE_STRING      UnicodeName;
    ULONG				ReadOffset;
	ULONG				ReadLength;
	ULONG				ReadBuffer;
	BOOLEAN				retval;

	fileNameInfo = (PFILE_NAME_INFORMATION) ExAllocatePool( NonPagedPool, 
                                                                MAXPATHLEN*sizeof(WCHAR) );
	RtlZeroMemory(fileNameInfo,MAXPATHLEN*sizeof(WCHAR));

	if (IsFastIO) {

		ReadOffset = ReadOffsetFastIO;
		ReadLength = Length;
		ReadBuffer = Buffer;

	} else {

		ReadOffset = curParameters->Parameters.Read.ByteOffset.LowPart;
		ReadLength = curParameters->Parameters.Read.Length;
		ReadBuffer = irp->UserBuffer;

	}

		// Ask at the fsd driver the filename from the fileobject
	    // this is 'legal' because we are inside an IRP_MJ_READ so the windows fsd
		// 'know' already this fileobject.

        if(fileNameInfo && 
            FFsdQueryFile(hookExt->FileSystem, fileObject, FileNameInformation, fileNameInfo, 
							(MAXPATHLEN - 1)*sizeof(WCHAR),FFsdQueryFileComplete,IRP_MJ_QUERY_INFORMATION)) {
 
            UnicodeName.Length = (SHORT) fileNameInfo->FileNameLength;
            UnicodeName.Buffer = fileNameInfo->FileName;
            if(NT_SUCCESS(RtlUnicodeStringToAnsiString(&fileName, &UnicodeName, TRUE))) {

				
				// if IRP_MJ_READ bpx is set we can look in sice at each IRP parameters.
				// I have used inlineasm to have these values directly in the registers.
				// EAX = file offset inside the file where fsd have to start reading.
				// EBX = pointer at the (unicode) file name of to read from.
				// ECX = length (bytes) to read from the file.
				// EDI = buffer for bytes read from the file.
				if (bpxEnabled) {

					__asm
					{
						mov ebx,[UnicodeName.Buffer]
						mov eax,ReadOffset
						mov ecx,ReadLength
						mov edi,ReadBuffer
						int 3
					}
				}

            }
        }
		
        if (fileNameInfo) ExFreePool(fileNameInfo);
		return TRUE;
		
}



//******************************************************************************
//
// FFsdHookRoutine routine - Here pass all IRPs directed to the driver hooked
//							 How u can see we hook only IRP_MJ_READ. It's enough
//							 to know which files are read and to jump easly from
//							 here, inside the InTether drivers code.
//
//******************************************************************************
NTSTATUS 
FFsdHookRoutine( 
    PDEVICE_OBJECT HookDevice, 
    IN PIRP Irp)
{
    PIO_STACK_LOCATION  currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    PIO_STACK_LOCATION  nextIrpStack    = IoGetNextIrpStackLocation(Irp);
    PFILE_OBJECT        FileObject;
    PHOOK_EXTENSION     hookExt;
    NTSTATUS			ns;
	BOOLEAN				ret;
	PFILE_STANDARD_INFORMATION fileStandardInfo;

    FileObject = currentIrpStack->FileObject;
	hookExt = HookDevice->DeviceExtension;

    if(currentIrpStack->MajorFunction == IRP_MJ_READ) {

		FileObject = currentIrpStack->FileObject;
		ret = FFsdGetPathName(FALSE, FileObject, hookExt, currentIrpStack, Irp, 0, 0, 0); 
	}
    
    *nextIrpStack = *currentIrpStack;
    IoSetCompletionRoutine(Irp, FFsdHooked, NULL, FALSE, FALSE, FALSE);
    return IoCallDriver(hookExt->FileSystem, Irp);
}


//*********************************************************************************
//
// FFsdDeviceRoutine routine - Here pass all IRP directed to our FFSD filter driver
//							   Because we have registered a FastIoDeviceControl
//							   routine, all IRP_MJ_DEVICE_CONTROL are routed there. 
//
//*********************************************************************************
NTSTATUS 
FFsdDeviceRoutine( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp)
{
    PIO_STACK_LOCATION  irpStack;
    PVOID               inputBuffer;
    PVOID               outputBuffer;
    ULONG               inputBufferLength;
    ULONG               outputBufferLength;
    ULONG               ioControlCode;

    Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
	irpStack = IoGetCurrentIrpStackLocation (Irp);
	inputBuffer        = Irp->AssociatedIrp.SystemBuffer;
    inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputBuffer       = Irp->AssociatedIrp.SystemBuffer;
    outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    ioControlCode      = irpStack->Parameters.DeviceIoControl.IoControlCode;

    switch (irpStack->MajorFunction) {

    case IRP_MJ_CREATE:

        break;

    case IRP_MJ_CLOSE:

        break;

    case IRP_MJ_DEVICE_CONTROL:

        if( Irp->MdlAddress ) {

            outputBuffer = MmGetSystemAddressForMdl(Irp->MdlAddress);
        }

        FFsdFastIoDeviceControl(irpStack->FileObject, TRUE, inputBuffer, inputBufferLength, 
                                    outputBuffer, outputBufferLength,
                                    ioControlCode, &Irp->IoStatus, DeviceObject );
        break;
    }

    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;   
}


//*******************************************************************************
//
// FFsdDispatch routine - This is the main dispatch routine of our FFsd driver
//						  Here IRP directed to our FFsd driver are routed to
//						  to FFsdDeviceRoutine. And the IRP for the hooked driver
//						  go to FFsdHookRoutine.
//
//*******************************************************************************
NTSTATUS 
FFsdDispatch( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp 
    )
{

	PHOOK_EXTENSION     Ext;

	Ext = DeviceObject->DeviceExtension;

    if( Ext->Type == GUIINTERFACE ) {

        return FFsdDeviceRoutine( DeviceObject, Irp );

    } else {

        return FFsdHookRoutine( DeviceObject, Irp );
    }
    
}




//*******************************************************************************
//
// DriverEntry routine - We make our FFsd driver 'visible' to our FFsdLoader.exe
//
//*******************************************************************************
NTSTATUS 
DriverEntry(
    IN PDRIVER_OBJECT DriverObject, 
    IN PUNICODE_STRING RegistryPath)
{
    
    
    WCHAR                   deviceNameBuffer[]  = L"\\Device\\FFsd";
    UNICODE_STRING          deviceNameUnicodeString;
    WCHAR                   deviceLinkBuffer[]  = L"\\DosDevices\\FFsd";
    UNICODE_STRING          deviceLinkUnicodeString;
    ULONG                   i;
	PDEVICE_OBJECT          ourDevice;
	NTSTATUS                ns;
	PHOOK_EXTENSION			ourExt;


    RtlInitUnicodeString (&deviceNameUnicodeString, deviceNameBuffer);

    ns = IoCreateDevice (DriverObject, sizeof(HOOK_EXTENSION),
                                &deviceNameUnicodeString,FFSD,0,TRUE,&ourDevice);

   if(NT_SUCCESS(ns)) {

		ourExt = ourDevice->DeviceExtension;
        ourExt -> Type = GUIINTERFACE;

		RtlInitUnicodeString (&deviceLinkUnicodeString, deviceLinkBuffer);
        ns = IoCreateSymbolicLink (&deviceLinkUnicodeString,
                                         &deviceNameUnicodeString);
        if(!NT_SUCCESS(ns)) {

            IoDeleteDevice(ourDevice);
            return ns;            
        }

        for(i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++) {

            DriverObject->MajorFunction[i] = FFsdDispatch;
        }

		DriverObject->FastIoDispatch = &FastIOHook;

    } else {

        return ns;
    }

    return STATUS_SUCCESS;
}


//*************************************************************
//
// FFsdFastIoReleaseForCcFlush
//
//*************************************************************    
NTSTATUS 
FFsdFastIoReleaseForCcFlush( 
    IN PFILE_OBJECT FileObject,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    NTSTATUS            retval = STATUS_NOT_IMPLEMENTED;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return STATUS_NOT_IMPLEMENTED;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, ReleaseForCcFlush)) {
        
        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseForCcFlush( 
            FileObject, hookExt->FileSystem);

    }
    return retval;
}


//*************************************************************
//
// FFsdFastIoCheckIfPossible
//
//*************************************************************
BOOLEAN  
FFsdFastIoCheckifPossible( 
    IN PFILE_OBJECT FileObject, 
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length, 
    IN BOOLEAN Wait, 
    IN ULONG LockKey, 
    IN BOOLEAN CheckForReadOperation,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN         retval = FALSE;
    PHOOK_EXTENSION hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoCheckIfPossible) ) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoCheckIfPossible( 
            FileObject, FileOffset, Length,
            Wait, LockKey, CheckForReadOperation, IoStatus, hookExt->FileSystem);
    }

    return retval;
}


//**************************************************************
//
// FFsdFastIoAcquireForCcFlush
//
//************************************************************** 
NTSTATUS 
FFsdFastIoAcquireForCcFlush( 
    IN PFILE_OBJECT FileObject,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    NTSTATUS            retval = STATUS_NOT_IMPLEMENTED;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return STATUS_NOT_IMPLEMENTED;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, AcquireForCcFlush)) {

       retval = hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireForCcFlush( 
            FileObject, hookExt->FileSystem);
        
    }
    return retval;
}


//**************************************************************
//
// FFsdFastIoReleaseForModWrite
//
//**************************************************************
NTSTATUS 
FFsdFastIoReleaseForModWrite( 
    IN PFILE_OBJECT FileObject,
    IN struct _ERESOURCE *ResourceToRelease,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    NTSTATUS            retval = STATUS_NOT_IMPLEMENTED;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return STATUS_NOT_IMPLEMENTED;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, ReleaseForModWrite)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseForModWrite( 
            FileObject,  ResourceToRelease, hookExt->FileSystem);

    }
    return retval;
}


//*************************************************************
//
// FFsdFastIoQueryOpen
//
//*************************************************************
BOOLEAN 
FFsdFastIoQueryOpen( 
    IN PIRP Irp,
    OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    PFILE_OBJECT        FileObject;
    PIO_STACK_LOCATION  currentIrpStack;
    PIO_STACK_LOCATION  nextIrpStack;
    

    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoQueryOpen)) {

        currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
        nextIrpStack    = IoGetNextIrpStackLocation(Irp);
        FileObject      = currentIrpStack->FileObject;

        *nextIrpStack = *currentIrpStack;
        nextIrpStack->DeviceObject = hookExt->FileSystem;
        IoSetNextIrpStackLocation(Irp);

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryOpen( 
            Irp, NetworkInformation, hookExt->FileSystem);

        Irp->CurrentLocation++;
        Irp->Tail.Overlay.CurrentStackLocation++;

    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoMdlWriteCompleteCompressed
//
//***************************************************************  
BOOLEAN 
FFsdFastIoMdlWriteCompleteCompressed( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset, 
    IN PMDL MdlChain, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension; 

    if( FASTIOPRESENT(hookExt, MdlWriteCompleteCompressed)) {

       retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlWriteCompleteCompressed( 
            FileObject, FileOffset, MdlChain, hookExt->FileSystem);

    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoMdlReadCompleteCompressed
//
//***************************************************************   
BOOLEAN 
FFsdFastIoMdlReadCompleteCompressed( 
    IN PFILE_OBJECT FileObject,
    IN PMDL MdlChain, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, MdlReadCompleteCompressed)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlReadCompleteCompressed( 
            FileObject, MdlChain, hookExt->FileSystem);

    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoWriteCompressed
//
//***************************************************************
BOOLEAN 
FFsdFastIoWriteCompressed( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length,
    IN ULONG LockKey, 
    OUT PVOID Buffer,
    OUT PMDL *MdlChain, 
    OUT PIO_STATUS_BLOCK IoStatus,
    OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
    IN ULONG CompressedDataInfoLength, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoWriteCompressed)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoWriteCompressed( 
            FileObject, FileOffset, Length, LockKey, Buffer, MdlChain, IoStatus,
            CompressedDataInfo, CompressedDataInfoLength, hookExt->FileSystem);

    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoReadCompressed
//
//***************************************************************   
BOOLEAN 
FFsdFastIoReadCompressed( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length,
    IN ULONG LockKey, 
    OUT PVOID Buffer,
    OUT PMDL *MdlChain, 
    OUT PIO_STATUS_BLOCK IoStatus,
    OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
    IN ULONG CompressedDataInfoLength, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoReadCompressed)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoReadCompressed( 
            FileObject, FileOffset, Length, LockKey, Buffer, MdlChain, IoStatus,
            CompressedDataInfo, CompressedDataInfoLength, hookExt->FileSystem);

    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoMdlWriteComplete
//
//***************************************************************   
BOOLEAN 
FFsdFastIoMdlWriteComplete( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset, 
    IN PMDL MdlChain, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
	if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, MdlWriteComplete)) { 

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlWriteComplete( 
            FileObject, FileOffset, MdlChain, hookExt->FileSystem);

    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoPrepareMdlWrite
//
//***************************************************************
BOOLEAN 
FFsdFastIoPrepareMdlWrite( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length,
    IN ULONG LockKey, 
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;
    IoStatus->Status      = STATUS_NOT_IMPLEMENTED;
    IoStatus->Information = 0;

    if( FASTIOPRESENT(hookExt, PrepareMdlWrite)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->PrepareMdlWrite( 
            FileObject, FileOffset, Length, LockKey, MdlChain, IoStatus, 
            hookExt->FileSystem);

    } 
    return retval;
}


//****************************************************************
//
// FFsdFastIoMdlReadComplete
//
//****************************************************************   
BOOLEAN 
FFsdFastIoMdlReadComplete( 
    IN PFILE_OBJECT FileObject,
    IN PMDL MdlChain, 
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, MdlReadComplete)) {

        retval = (BOOLEAN) hookExt->FileSystem->DriverObject->FastIoDispatch->MdlReadComplete(FileObject, 
                                                                                               MdlChain, hookExt->FileSystem);
    }
    return retval;
}


//***************************************************************
//
// FFsdFastIoMdlRead
//
//***************************************************************  
BOOLEAN 
FFsdFastIoMdlRead( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length,
    IN ULONG LockKey, 
    OUT PMDL *MdlChain,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, MdlRead)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->MdlRead( 
            FileObject, FileOffset, Length, LockKey, MdlChain, 
            IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//**************************************************************
//
// FFsdFastIoAcquireForModWrite
//
//**************************************************************    
NTSTATUS 
FFsdFastIoAcquireForModWrite( 
    IN PFILE_OBJECT FileObject,
    IN PLARGE_INTEGER EndingOffset,
    OUT struct _ERESOURCE **ResourceToRelease,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    NTSTATUS            retval = STATUS_NOT_IMPLEMENTED;
    PHOOK_EXTENSION     hookExt;

    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, AcquireForModWrite)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireForModWrite( 
            FileObject, EndingOffset, ResourceToRelease, hookExt->FileSystem);

    }
    return retval;
}


//*************************************************************
//
// FFsdFastIoQueryNetworkOpenInfo
//
//*************************************************************   
BOOLEAN 
FFsdFastIoQueryNetworkOpenInfo(
    IN PFILE_OBJECT FileObject,
    IN BOOLEAN Wait,
    OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
    OUT PIO_STATUS_BLOCK IoStatus,
    IN PDEVICE_OBJECT DeviceObject 
    )
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    if( !DeviceObject ) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoQueryNetworkOpenInfo)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryNetworkOpenInfo( 
            FileObject, Wait, Buffer, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//**************************************************************
//
// FFsdFastIoDetachDevice
//
// *************************************************************
VOID 
FFsdFastIoDetachDevice( 
    PDEVICE_OBJECT SourceDevice, 
    PDEVICE_OBJECT TargetDevice 
    ) 
{
    PHOOK_EXTENSION     hookExt;
    ULONG               i;
    
    hookExt = SourceDevice->DeviceExtension;
    if(FASTIOPRESENT( hookExt, FastIoDetachDevice)) {

        hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoDetachDevice( 
            SourceDevice, TargetDevice);
    }
}


//**************************************************************
//
// FFsdFastIoReleaseFile
//
//**************************************************************
VOID 
FFsdFastIoReleaseFile( 
    PFILE_OBJECT FileObject 
    ) 
{
    PDEVICE_OBJECT      deviceObject, checkDevice;
    PHOOK_EXTENSION     hookExt;
    
    checkDevice = FileObject->DeviceObject->Vpb->DeviceObject;
    while( checkDevice ) {

        if( checkDevice->DriverObject == FFsdDriver) {
    
            deviceObject = IoGetRelatedDeviceObject( FileObject );
            hookExt = deviceObject->DeviceExtension;
            
            if( FASTIOPRESENT(hookExt, ReleaseFileForNtCreateSection)) {

                 hookExt->FileSystem->DriverObject->FastIoDispatch->ReleaseFileForNtCreateSection(FileObject);
            }
            return;
        }
        checkDevice = checkDevice->AttachedDevice;
    }
}


//**************************************************************
//
// FFsdFastIoAcquireFile
//
//**************************************************************
VOID 
FFsdFastIoAcquireFile( 
    PFILE_OBJECT FileObject 
    ) 
{
    PDEVICE_OBJECT      deviceObject, checkDevice;
    PHOOK_EXTENSION     hookExt;
    
    checkDevice = FileObject->DeviceObject->Vpb->DeviceObject;
    while(checkDevice) {

        if(checkDevice->DriverObject == FFsdDriver) {
    
            deviceObject = checkDevice;
            hookExt = deviceObject->DeviceExtension;
            if( FASTIOPRESENT( hookExt, AcquireFileForNtCreateSection)) {

                hookExt->FileSystem->DriverObject->FastIoDispatch->AcquireFileForNtCreateSection(FileObject);

            }
            return;
        }
        checkDevice = checkDevice->AttachedDevice;
    }
}

//*******************************************************************************
//
// FFsdHookDrive routine - we hook the 'C' driver. If u want hook all or others
//						   drives u have to change this routine. But keep in
//						   mind that this driver is just a simple filter coded to
//						   reverse InTether protection and It's not a full
//						   fs filter driver.
//
//*******************************************************************************
BOOLEAN 
FFsdHookDrive( 
    IN ULONG Drive, 
    IN PDRIVER_OBJECT DriverObject 
    )
{
    ULONG               i;
    PDEVICE_OBJECT      hookDevice;
    UNICODE_STRING      fileNameUnicodeString;
    PFILE_FS_ATTRIBUTE_INFORMATION fileFsAttributes;
    WCHAR               filename[] = L"\\DosDevices\\C:\\";
    NTSTATUS            ns;
	IO_STATUS_BLOCK     ioStatus;
    HANDLE              FileHandle;   
    OBJECT_ATTRIBUTES   objectAttributes;
    PDEVICE_OBJECT      fileSysDevice;
    PFILE_OBJECT        fileObject;
    PHOOK_EXTENSION     hookExtension;
	ULONG               fileFsAttributesSize;
    
    
	if(DriveHookDevice == NULL)  {

        filename[12] = (CHAR) ('C'+Drive);
		
        RtlInitUnicodeString(&fileNameUnicodeString, filename);
        InitializeObjectAttributes(&objectAttributes, &fileNameUnicodeString, 
                                    OBJ_CASE_INSENSITIVE, NULL, NULL);
        ns = ZwCreateFile(&FileHandle, SYNCHRONIZE|FILE_ANY_ACCESS, 
                                 &objectAttributes, &ioStatus, NULL, 0, FILE_SHARE_READ|FILE_SHARE_WRITE, 
                                 FILE_OPEN, 
                                 FILE_SYNCHRONOUS_IO_NONALERT|FILE_DIRECTORY_FILE, 
                                 NULL, 0);
        if(!NT_SUCCESS(ns)) {

            return FALSE;
        }

        ns = ObReferenceObjectByHandle(FileHandle, FILE_READ_DATA, 
                                              NULL, KernelMode, &fileObject, NULL);
        if(!NT_SUCCESS(ns)) {

            ZwClose(FileHandle);
            return FALSE;
        }

        fileSysDevice = IoGetRelatedDeviceObject(fileObject);

        if(!fileSysDevice) {

            ObDereferenceObject(fileObject);
            ZwClose(FileHandle);
            return FALSE;
        }

        ns = IoCreateDevice(DriverObject,sizeof(HOOK_EXTENSION),NULL,
                                   fileSysDevice->DeviceType,0,
                                   FALSE,
                                   &hookDevice );
        if(!NT_SUCCESS(ns)) {

            ObDereferenceObject(fileObject);
            ZwClose(FileHandle);

            return FALSE;
        }

        hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;

        hookExtension = hookDevice->DeviceExtension;
        hookExtension->LogicalDrive = 'C'+Drive;
        hookExtension->FileSystem   = fileSysDevice;
        hookExtension->Hooked       = TRUE;
        hookExtension->Type         = STANDARD;

        ns = IoAttachDeviceByPointer(hookDevice, fileSysDevice);
        if(!NT_SUCCESS(ns))  {

            ObDereferenceObject(fileObject);
            ZwClose(FileHandle);

            return FALSE;

        } else {

            
        }

       fileFsAttributesSize = sizeof( FILE_FS_ATTRIBUTE_INFORMATION) + MAXPATHLEN;
        hookExtension->FsAttributes = (PFILE_FS_ATTRIBUTE_INFORMATION) ExAllocatePool( NonPagedPool, 
                                                                                       fileFsAttributesSize );
        if( hookExtension->FsAttributes &&
            !NT_SUCCESS( IoQueryVolumeInformation( fileObject, FileFsAttributeInformation,
                                                   fileFsAttributesSize, hookExtension->FsAttributes, 
                                                   &fileFsAttributesSize ))) {

            ExFreePool( hookExtension->FsAttributes );
            hookExtension->FsAttributes = NULL;
        } 

        ObDereferenceObject(fileObject);

        ZwClose(FileHandle);

        DriveHookDevice = hookDevice;
        
    } else {

        hookExtension = DriveHookDevice->DeviceExtension;
        hookExtension->Hooked = TRUE;
    }
    return TRUE;
}



//***************************************************************************
//
// FFSdFastIoDeviceControl - here we manage the IRP_IO_DEVICE_CONTROL for our
//							 FFsd driver. Basically only 2 ;)
//
//***************************************************************************
BOOLEAN  
FFsdFastIoDeviceControl( 
    IN PFILE_OBJECT FileObject, 
    IN BOOLEAN Wait,
    IN PVOID InputBuffer, 
    IN ULONG InputBufferLength, 
    OUT PVOID OutputBuffer, 
    IN ULONG OutputBufferLength, 
    IN ULONG IoControlCode,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    hookExt = DeviceObject->DeviceExtension;
    if(hookExt->Type == GUIINTERFACE) {

        IoStatus->Status = STATUS_SUCCESS;
        IoStatus->Information = 0;

        switch (IoControlCode) {

		case IOCTL_FFSD_STARTHOOK:

            retval = FFsdHookDrive(0, DeviceObject->DriverObject);
            IoStatus->Information = sizeof(ULONG);
            break;

		case IOCTL_FFSD_ENABLEBPX:

			if(!bpxEnabled){

				bpxEnabled = TRUE;

			}else{
				
				bpxEnabled = FALSE;
			}
			IoStatus->Information = sizeof(ULONG);
			break;

default:

            IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
            break;
        }

        retval = TRUE;

    } else {
    
		if( FASTIOPRESENT(hookExt, FastIoDeviceControl)) {
        
			retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoDeviceControl( 
                FileObject, Wait, InputBuffer, InputBufferLength, OutputBuffer, 
                OutputBufferLength, IoControlCode, IoStatus, hookExt->FileSystem);

        }
	}

    return retval;
}


//************************************************************
//
// FFsdFastIoUnlockAllByKey
//
//************************************************************
BOOLEAN  
FFsdFastIoUnlockAllByKey( 
    IN PFILE_OBJECT FileObject, 
    PEPROCESS ProcessId, 
    ULONG Key,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoUnlockAllByKey)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockAllByKey( 
            FileObject, ProcessId, Key, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//***********************************************************
//
// FFsdFastIoUnlockAll
//
//***********************************************************
BOOLEAN  
FFsdFastIoUnlockAll( 
    IN PFILE_OBJECT FileObject, 
    PEPROCESS ProcessId,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if(FASTIOPRESENT(hookExt, FastIoUnlockAll)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockAll( 
            FileObject, ProcessId, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//**********************************************************
//
// FFsdFastIoUnlockSingle
//
//**********************************************************
BOOLEAN  
FFsdFastIoUnlockSingle( 
    IN PFILE_OBJECT FileObject, 
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length, 
    PEPROCESS ProcessId, 
    ULONG Key,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;   
    
    if( FASTIOPRESENT(hookExt, FastIoUnlockSingle)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoUnlockSingle(
            FileObject, FileOffset, Length, ProcessId, Key, 
            IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//**********************************************************
//
// FFsdFastIoLock
//
//**********************************************************
BOOLEAN  
FFsdFastIoLock( 
    IN PFILE_OBJECT FileObject, 
    IN PLARGE_INTEGER FileOffset,
    IN PLARGE_INTEGER Length, 
    PEPROCESS ProcessId, 
    ULONG Key,
    BOOLEAN FailImmediately, 
    BOOLEAN ExclusiveLock,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if(FASTIOPRESENT(hookExt, FastIoLock)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoLock( 
            FileObject, FileOffset, Length, ProcessId, Key, FailImmediately, 
            ExclusiveLock, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//*********************************************************
//
// FFsdFastIoQueryStandardInfo
//
//*********************************************************
BOOLEAN  
FFsdFastIoQueryStandardInfo( 
    IN PFILE_OBJECT FileObject, 
    IN BOOLEAN Wait, 
    OUT PFILE_STANDARD_INFORMATION Buffer,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoQueryStandardInfo)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryStandardInfo( 
            FileObject, Wait, Buffer, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//********************************************************
//
// FFsdFastIoQueryBasicinfo
//
//********************************************************
BOOLEAN  
FFsdFastIoQueryBasicInfo( 
    IN PFILE_OBJECT FileObject, 
    IN BOOLEAN Wait, 
    OUT PFILE_BASIC_INFORMATION Buffer,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;

    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoQueryBasicInfo)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoQueryBasicInfo( 
            FileObject, Wait, Buffer, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//********************************************************
//
// FFsdFastIoWrite
//
//********************************************************
BOOLEAN  
FFsdFastIoWrite( 
    IN PFILE_OBJECT FileObject, 
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length, 
    IN BOOLEAN Wait, 
    IN ULONG LockKey, 
    IN PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject 
    ) 
{
    BOOLEAN              retval = FALSE;
    PHOOK_EXTENSION      hookExt;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoWrite)) {

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoWrite( 
            FileObject, FileOffset, Length, Wait, LockKey, 
            Buffer, IoStatus, hookExt->FileSystem);

    }
    return retval;
}


//********************************************************
// 
// FFsdFastIoRead
//
//********************************************************
BOOLEAN  
FFsdFastIoRead( 
    IN PFILE_OBJECT FileObject, 
    IN PLARGE_INTEGER FileOffset, 
    IN ULONG Length, 
    IN BOOLEAN Wait, 
    IN ULONG LockKey, 
    OUT PVOID Buffer,
    OUT PIO_STATUS_BLOCK IoStatus, 
    IN PDEVICE_OBJECT DeviceObject
    ) 
{
    BOOLEAN             retval = FALSE;
    PHOOK_EXTENSION     hookExt;
	BOOLEAN				ret;
    
    if(!DeviceObject) return FALSE;

    hookExt = DeviceObject->DeviceExtension;

    if( FASTIOPRESENT(hookExt, FastIoRead)) {

		ret = FFsdGetPathName(TRUE, FileObject, hookExt, 0, 0, FileOffset->LowPart, Length, Buffer); 

        retval = hookExt->FileSystem->DriverObject->FastIoDispatch->FastIoRead( 
            FileObject, FileOffset, Length,
            Wait, LockKey, Buffer, IoStatus, hookExt->FileSystem);

    }
    return retval;
}