
#include "ntddk.h"
#include "intetkey.h"
#include "stdio.h"



// =================================================================
// GLOBAL DATA
// =================================================================

PDEVICE_OBJECT  IKDeviceObject;




typedef struct _FILE_INTERNAL_INFORMATION {
	LARGE_INTEGER FileId;
} FILE_INTERNAL_INFORMATION, *PFILE_INTERNAL_INFORMATION;




//***************************************************************************
//
// IKGetKey - Read the 2047 bytes key in dgcrypt data section in memory
//
//***************************************************************************
NTSTATUS
IKGetKey (ULONG OutputBuffer)
    {
    
    NTSTATUS	ns = STATUS_INVALID_PARAMETER;

	ULONG i, n, *q;
	PSYSTEM_MODULE_INFORMATION p;
	ULONG *dgcrypt = 0;


	ZwQuerySystemInformation (SystemModuleInformation, &n, 0, &n);
	q = (ULONG *) ExAllocatePool (PagedPool, n);
	ZwQuerySystemInformation (SystemModuleInformation, q, n * sizeof (*q), 0);

	p = (PSYSTEM_MODULE_INFORMATION) ( q + 1 );
	dgcrypt = 0;
	for ( i = 0; i < *q; ++ i )

		if ( _stricmp( p[i].abPath + p[i].wNameOffset,"dgcrypt.sys") == 0 ) {

			dgcrypt = p[i].pBase;
			break;
		}

	ExFreePool (q);
	if (dgcrypt) {

		// point to the start of the key buffer used by InTether to encrypt/decrypt
		// our text files. But keep in mind that the data section of dgcrypt.sys
		// is full of keys used for other types of encryption. So you have to write
		// your own routine here depend of what key you want to dump.
		// For this tutorial (part1) it's enough this to give an idea ;)
		__asm 
		{
			add  dword ptr [dgcrypt],0000A980h
		}
		
		RtlCopyMemory (OutputBuffer,dgcrypt,0x7FF);

	}
	


	ns = STATUS_SUCCESS;
	return ns;

    }


//***************************************************************************
//
// IKIoDeviceControl - here we manage the IRP_IO_DEVICE_CONTROL 
//
//***************************************************************************
BOOLEAN  
IKIoDeviceControl( 
    IN PVOID InputBuffer, 
    IN ULONG InputBufferLength, 
    OUT PVOID OutputBuffer, 
    IN ULONG OutputBufferLength, 
    IN ULONG IoControlCode,
    OUT PIO_STATUS_BLOCK IoStatus) 
{
    BOOLEAN             retval = FALSE;
    

        IoStatus->Status = STATUS_SUCCESS;
        IoStatus->Information = 0;

        switch (IoControlCode) {

		case IOCTL_INTETKEY:

            retval = IKGetKey(OutputBuffer);
            IoStatus->Information = 0x000007FF;
            break;

default:

            IoStatus->Status = STATUS_INVALID_DEVICE_REQUEST;
            break;
        }

    retval = TRUE;
    return retval;
}


//*******************************************************************************
//
// IKDeviceDispatcher routine - We take care of IRP_MJ_DEVICE_CONTROL
//
//*******************************************************************************
NTSTATUS 
IKDispatcher (POUREXTENSION pDeviceext,
							 PIRP Irp)
    {

    PIO_STACK_LOCATION  irpStack;
    PVOID               inputBuffer;
    PVOID               outputBuffer;
    ULONG               inputBufferLength;
    ULONG               outputBufferLength;
    ULONG               ioControlCode;
	NTSTATUS            ns;

	Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
	irpStack = IoGetCurrentIrpStackLocation (Irp);

	inputBuffer        = Irp->AssociatedIrp.SystemBuffer;
    inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
    outputBuffer       = Irp->AssociatedIrp.SystemBuffer;
    outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    ioControlCode      = irpStack->Parameters.DeviceIoControl.IoControlCode;

    switch (irpStack->MajorFunction) {

        case IRP_MJ_CREATE:

        case IRP_MJ_CLEANUP:

        case IRP_MJ_CLOSE:

            {
            ns = STATUS_SUCCESS;
            break;
            }

        case IRP_MJ_DEVICE_CONTROL:

            {
            ns = IKIoDeviceControl (inputBuffer, inputBufferLength, outputBuffer, outputBufferLength,
                                    ioControlCode, &Irp->IoStatus);
            break;
            }
        }
    
	IoCompleteRequest (Irp, IO_NO_INCREMENT);
    return ns;
    }







//*******************************************************************************
//
// DriverUnload routine - We delete InTetKey symbolic link
//
//*******************************************************************************
void
DriverUnload (PDRIVER_OBJECT DriverObject)
    {

	UNICODE_STRING  usSymbolicLinkName;

	RtlInitUnicodeString(&usSymbolicLinkName, L"\\??\\Intetkey");
    IoDeleteSymbolicLink (&usSymbolicLinkName);
    IoDeleteDevice       (IKDeviceObject);
    return;
    }



//*******************************************************************************
//
// DriverEntry routine - We make InTetKey driver 'visible' to our InTetBackup.exe
//
//*******************************************************************************
NTSTATUS 
DriverEntry(
    IN PDRIVER_OBJECT DriverObject, 
    IN PUNICODE_STRING RegistryPath)
{
    
    
    WCHAR                   deviceNameBuffer[]  = L"\\Device\\Intetkey";
    UNICODE_STRING          deviceNameUnicodeString;
    WCHAR                   deviceLinkBuffer[]  = L"\\DosDevices\\Intetkey";
    UNICODE_STRING          deviceLinkUnicodeString;
    ULONG                   i;
	PDEVICE_OBJECT          ourDevice;
	NTSTATUS                ns;
	POUREXTENSION			ourExt;

    RtlInitUnicodeString (&deviceNameUnicodeString, deviceNameBuffer);

    ns = IoCreateDevice (DriverObject, sizeof(OUREXTENSION),
                                &deviceNameUnicodeString,0,0,TRUE,&ourDevice);

   if(NT_SUCCESS(ns)) {

		RtlInitUnicodeString (&deviceLinkUnicodeString, deviceLinkBuffer);
        ns = IoCreateSymbolicLink (&deviceLinkUnicodeString,
                                         &deviceNameUnicodeString);
        if(!NT_SUCCESS(ns)) {

            IoDeleteDevice(ourDevice);
            return ns;            
        }

		IKDeviceObject = ourDevice;
		ourExt = ourDevice->DeviceExtension;
		ourExt->DriverObject = DriverObject;
        ourExt->DeviceObject = ourDevice;

        for(i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++) {

            DriverObject->MajorFunction[i] = IKDispatcher;
        }

		DriverObject->DriverUnload = DriverUnload;

    } else {

        return ns;
    }

    return STATUS_SUCCESS;
}






