
BOOTSECTOR STRUCT
  Jump			db 3 dup (?)		; 3 bytes jump opcode
  Format  		db 8 dup (?)		; 8 bytes ascii to describe the filesystem format 
  BytesPerSector	dw ?			; Number of bytes for a sector
  SectorsPerCluster	db ?			; Number of sectors for a cluster
  BootSectors		dw ?
  Mbz1			db ?
  Mbz2			dw ?
  Reserved1		dw ?
  MediaType		db ?
  Mbz3			dw ?
  SectorsPerTrack	dw ?
  NumberOfHeads		dw ?
  PartitionOffset	dd ?
  Reserved2		dd 2 dup (?)
  TotalSectors		dd 2 dup (?)
  MftStartLcn		dd 2 dup (?)		; Start Lcn (Logical cluster number) of the MFT
  MftStart2StartLcn	dd 2 dup (?)		; a copy of previous value for security reason
  ClustersPerFileRecord	dd ?			; Number of cluster for file record (MFT entry)
  ClustersPerIndexBlock	dd ?
  VolumeSerialNumber	dd 2 dup (?)
  BootCode		db 431 dup (?)
  BootSignature		dw ?
BOOTSECTOR ENDS


NTFS_RECORD_HEADER STRUCT
  RecordType		dd ?			; Type of FileRecord 'FILE', 'INDX', 'BAAD', 'HOLE', 'CHKD'
  UsaOffset		dw ?			; Offset from the start of this structure to the Update Sequence Array
  UsaCount		dw ?			; Number of value in the Update Sequence Array
  Usn			dd 2 dup (?)		; Record's Update Sequence Number 
NTFS_RECORD_HEADER ENDS


FILE_RECORD_HEADER STRUCT
  NTFS_RecordHeader	NTFS_RECORD_HEADER <>
  SequenceNumber	dw ?			; Number of time this MFT entry has been used
  LinkCount		dw ?			; Number of directory link at this MFT entry
  AttributeOffset	dw ?			; Offset from the start of this structure to the first Attribute of this MFT entry
  Flags			dw ?			; Bits flag :
  						; 0001 = This FileRecord in use
  						; 0002 = This FileRecord is a directory
  BytesInUse		dd ?			; Number of bytes used by this MFT entry
  BytesAllocated	dd ?			; Number of bytes allocated for this MFT entry
  BaseFileRecord	dd 2 dup (?)		; If the filerecord use Attribute longer than the default size this value store a reference to the file
  NextAttributeNumber	dw ?			; Number that will be used for a new Attribute
FILE_RECORD_HEADER ENDS


ATTRIBUTE STRUCT
  AttributeType		dd ?			; Attribute Type (see tutorial)
  LengthResAtt		dd ?			; Size in bytes of the resident part of the Attribute
  NonResident		db ?			; 0 = Attribute resident, 1 = Attribute Nonresident
  NameLength		db ?			; Size in chars of the Attribute name (if exist)
  NameOffset		dw ?			; Offset from the start of this structure to the the Attribute name
  Flags			dw ?			; Bits flag 
  						; 0001 = compressed
  AttributeNumber	dw ?			; ID for the Attribute
ATTRIBUTE ENDS


RESIDENT_ATTRIBUTE STRUCT
  ValueLength		dd ?			; Size in bytes of the Attribute value
  ValueOffset		dw ?			; Offset from the start of this structure to Attribute value
  Flags			dw ?			; Bit flag for attribute propriety
RESIDENT_ATTRIBUTE ENDS


NONRESIDENT_ATTRIBUTE STRUCT
  LowVCN		dd 2 dup (?)		; The lowest Vcn (virtual cluster number) valid
  HighVCN		dd 2 dup (?)		; The highest Vcn (virtual cluster number) valid
  RunAwayOffset		dw ?			; Offset form the start of this structure to the RunAway array
  CompressionUnit	db ?			; Compression unit. If 0 the attribute is not compressed
  Reserved		db 5 dup (?)		; absolutaly no idea :-D
  AllocatedSize		dd 2 dup (?)		; Allocated space on disk to store the attribute
  DataSize		dd 2 dup (?)		; Size of the attribute (be carefull for compressed attribute!)
  InitializedSize	dd 2 dup (?)		; Size of the init part of the attribute
  CompressedSize	dd 2 dup (?)		; compressed size of the attribute (if the attribute is not compressed this entry doesn't exist)
NONRESIDENT_ATTRIBUTE ENDS



ATTRIBUTE_FILE_NAME STRUCT
  DirectoryFileRef	dd 2 dup (?)		 
  CreationTime		dd 2 dup (?)	
  ChangeTime		dd 2 dup (?)
  LastWriteTime		dd 2 dup (?)
  LastAccessTime	dd 2 dup (?)
  AllocatedSize		dd 2 dup (?)
  DataSize		dd 2 dup (?)
  FileAttributes	dd ?
  Reserved		dd ?
  NameLength		db ?
  NameType		db ?
  FileName		db 255 dup (?)
ATTRIBUTE_FILE_NAME ENDS


ATTRIBUTE_STANDARD_INFORMATION STRUCT
  CreateTime		dd 2 dup (?)
  ChangeTime		dd 2 dup (?)
  LastWriteTime		dd 2 dup (?)
  LastAccessTime	dd 2 dup (?)
  FileAttributes	dd ?
  Reserved		dd 3 dup (?)
  QuotaId		dd ?
  SecurityId		dd ?
  QuotaCharge		dd 2 dup (?)
  Usn			dd 2 dup (?)
ATTRIBUTE_STANDARD_INFORMATION ENDS